# -*- coding: utf-8 -*-
import json
import sys
from time import time

"""
Slightly modified wox.py credit: https://github.com/Wox-launcher/Wox
"""

class Launcher(object):
    """
    Launcher python plugin base
    """

    def __init__(self, API):
        self.rpc_request = {'method': 'query', 'parameters': ['']}
        if len(sys.argv) > 1:
            self.rpc_request = json.loads(sys.argv[1])

        if 'settings' in self.rpc_request.keys():
            self._settings = self.rpc_request['settings']
        if not self._debug:
            self._debug = self.settings.get('debug', False)
        if self._debug:
            self.logger_level("debug")
        self.logger.debug(f'Request:\n{json.dumps(self.rpc_request, indent=4)}')
        self.logger.debug(f"Params: {self.rpc_request.get('parameters')}")

    def __call__(self):
        # proxy is not working now
        # self.proxy = rpc_request.get("proxy",{})
        request_method_name = self.rpc_request.get("method")
        #transform query and context calls to internal flox methods
        if request_method_name == 'query' or request_method_name == 'context_menu':
            request_method_name = f"_{request_method_name}"

        request_parameters = self.rpc_request.get("parameters")

        request_method = getattr(self, request_method_name)
        try:
            results = request_method(*request_parameters)
        except Exception as e:
            try:
                self.logger.exception(f'Exception while calling method: {request_method_name}')
            except AttributeError:
                pass
            raise
        line_break = '#' * 10
        ms = int((time() - self._start) * 1000)
        self.logger.debug(f'{line_break} Total time: {ms}ms {line_break}')
        if request_method_name == "_query" or request_method_name == "_context_menu":
            results = {"result": results}
            if self._settings != self.rpc_request.get('Settings') and self._settings is not None:
                results['SettingsChange'] = self._settings

            print(json.dumps(results))

    def __del__(self):
        self.__call__()

    def query(self,query):
        """
        sub class need to override this method
        """
        return []

    def context_menu(self, data):
        """
        optional context menu entries for a result
        """
        return []

    def debug(self,msg):
        """
        alert msg
        """
        print("DEBUG:{}".format(msg))
        sys.exit()

    def change_query(self, query, requery=False):
        """
        change query
        """
        print(json.dumps({"method": f"{self.api}.ChangeQuery","parameters":[query,requery]}))

    def shell_run(self, cmd):
        """
        run shell commands
        """
        print(json.dumps({"method": f"{self.api}.ShellRun","parameters":[cmd]}))

    def close_app(self):
        """
        close launcher
        """
        print(json.dumps({"method": f"{self.api}.CloseApp","parameters":[]}))

    def hide_app(self):
        """
        hide launcher
        """
        print(json.dumps({"method": f"{self.api}.HideApp","parameters":[]}))

    def show_app(self):
        """
        show launcher
        """
        print(json.dumps({"method": f"{self.api}.ShowApp","parameters":[]}))

    def show_msg(self, title, sub_title, ico_path=""):
        """
        show messagebox
        """
        print(json.dumps({"method": f"{self.api}.ShowMsg","parameters":[title,sub_title,ico_path]}))

    def open_setting_dialog(self):
        """
        open setting dialog
        """
        self.logger.debug(json.dumps({"method": f"{self.api}.OpenSettingDialog","parameters":[]}))
        print(json.dumps({"method": f"{self.api}.OpenSettingDialog","parameters":[]}))

    def start_loadingbar(self):
        """
        start loading animation in wox
        """
        print(json.dumps({"method": f"{self.api}.StartLoadingBar","parameters":[]}))

    def stop_loadingbar(self):
        """
        stop loading animation in wox
        """
        print(json.dumps({"method": f"{self.api}.StopLoadingBar","parameters":[]}))

    def reload_plugins(self):
        """
        reload all launcher plugins
        """
        print(json.dumps({"method": f"{self.api}.ReloadPlugins","parameters":[]}))



from tempfile import gettempdir
from urllib import request
from urllib.error import URLError
from pathlib import Path
from functools import wraps
import json
import os
from time import time
import socket
from concurrent.futures import ThreadPoolExecutor
import logging

logging = logging.getLogger(__name__)

URL_SCHEMA = [
    'http://',
    'https://',
]
socket.setdefaulttimeout(15)

def cache(file_name:str, max_age=30, dir=gettempdir()):
    """
    Cache decorator
    """
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            cache_file = Path(dir, file_name)
            if cache_file.exists() and file_age(cache_file) < max_age and cache_file.stat().st_size != 0:
                with open(cache_file, 'r', encoding='utf-8') as f:
                    try:
                        cache = json.load(f)
                    except json.JSONDecodeError:
                        logging.warning('Unable to read cache file: %s', cache_file)
                        f.close()
                        os.remove(cache_file)
                    else:
                        return cache
            data = func(*args, **kwargs)
            if data is None:
                return None
            if len(data) != 0:
                try:
                    write_json(data, cache_file)
                except FileNotFoundError:
                    logging.warning('Unable to write cache file: %s', cache_file)
            return data
        return wrapper
    return decorator

def read_json(path:str):
    """
    Read json file
    """
    with open(path, 'r', encoding='utf-8') as f:
        data = json.load(f)

def write_json(data, path):
    if not Path(path).parent.exists():
        Path(path).parent.mkdir(parents=True)
    with open(path, 'w') as f:
        json.dump(data, f)

def file_age(path):
    age = time() - path.stat().st_mtime
    return age

def get_cache(path, max_age=0):
    if Path(path).exists() and file_age(path) < max_age and path.stat().st_size != 0:
        return read_json(path)
    return None

def refresh_cache(file_name:str, dir:str=gettempdir()):
    """
    Touch cache file
    """
    cache_file = Path(dir, file_name)
    if cache_file.exists():
        cache_file.touch()

def cache_path(file_name:str, dir:str=gettempdir()):
    """
    Return path to cache file
    """
    return Path(dir, file_name)

def remove_cache(file_name:str, dir:str=gettempdir()):
    """
    Remove cache file
    """
    cache_file = Path(dir, file_name)
    if cache_file.exists():
        cache_file.unlink()

def download_file(url:str, path, **kwargs):
    """
    Download file from url and save it to dir

    Args:
        url (str): image url.
        dir (str): directory to save image.
        file_name (str): file name to save image.

    Keyword Args:
        force_download (bool): Force download image even if it exists.
    """
    force_download = kwargs.pop('force_download', False)
    if not force_download and path.exists():
        return
    try:
        request.urlretrieve(url, path)
    except URLError as e:
        logging.exception(f'Unable to download: {url}')
    return Path(path)

def get_icon(url:str, path, file_name:str=None, **kwargs):
    for schema in URL_SCHEMA:
        if url.startswith(schema):
            break
    else:
        return url
    executor = kwargs.pop('executor', False)
    if file_name is None:
        file_name = url.split('/')[-1]
    if not Path(path).is_absolute():
        path = Path(gettempdir(), path)
    if not path.exists():
        path.mkdir(parents=True)
    full_path = Path(path, file_name)
    if not full_path.exists():
        if executor is False:
            download_file(url, full_path)
        else:
            executor.submit(download_file, url, full_path)
    return full_path


import ctypes

from ctypes.wintypes import BOOL, HWND, HANDLE, HGLOBAL, UINT, LPVOID
from ctypes import c_size_t as SIZE_T

# Credit for code goes to Mark Ransom at https://stackoverflow.com/a/25678113

OpenClipboard = ctypes.windll.user32.OpenClipboard
OpenClipboard.argtypes = HWND,
OpenClipboard.restype = BOOL
EmptyClipboard = ctypes.windll.user32.EmptyClipboard
EmptyClipboard.restype = BOOL
GetClipboardData = ctypes.windll.user32.GetClipboardData
GetClipboardData.argtypes = UINT,
GetClipboardData.restype = HANDLE
SetClipboardData = ctypes.windll.user32.SetClipboardData
SetClipboardData.argtypes = UINT, HANDLE
SetClipboardData.restype = HANDLE
CloseClipboard = ctypes.windll.user32.CloseClipboard
CloseClipboard.restype = BOOL
CF_UNICODETEXT = 13

GlobalAlloc = ctypes.windll.kernel32.GlobalAlloc
GlobalAlloc.argtypes = UINT, SIZE_T
GlobalAlloc.restype = HGLOBAL
GlobalLock = ctypes.windll.kernel32.GlobalLock
GlobalLock.argtypes = HGLOBAL,
GlobalLock.restype = LPVOID
GlobalUnlock = ctypes.windll.kernel32.GlobalUnlock
GlobalUnlock.argtypes = HGLOBAL,
GlobalSize = ctypes.windll.kernel32.GlobalSize
GlobalSize.argtypes = HGLOBAL,
GlobalSize.restype = SIZE_T

GMEM_MOVEABLE = 0x0002
GMEM_ZEROINIT = 0x0040

unicode_type = type(u'')

class Clipboard(object):

    def get(self):
        return get()

    def put(self, text):
        return put(text)

def get():
    """
    Get the contents of the clipboard.
    """
    text = None
    OpenClipboard(None)
    handle = GetClipboardData(CF_UNICODETEXT)
    pcontents = GlobalLock(handle)
    size = GlobalSize(handle)
    if pcontents and size:
        raw_data = ctypes.create_string_buffer(size)
        ctypes.memmove(raw_data, pcontents, size)
        text = raw_data.raw.decode('utf-16le').rstrip(u'\0')
    GlobalUnlock(handle)
    CloseClipboard()
    return text

def put(s):
    """
    Put the given string onto the clipboard.
    """
    if not isinstance(s, unicode_type):
        s = s.decode('mbcs')
    data = s.encode('utf-16le')
    OpenClipboard(None)
    EmptyClipboard()
    handle = GlobalAlloc(GMEM_MOVEABLE | GMEM_ZEROINIT, len(data) + 2)
    pcontents = GlobalLock(handle)
    ctypes.memmove(pcontents, data, len(data))
    GlobalUnlock(handle)
    SetClipboardData(CF_UNICODETEXT, handle)
    CloseClipboard()

def copy(s):
    put(s)